import json
import time
from datetime import UTC, datetime, timedelta
from typing import Literal, NamedTuple

import boto3

from rds_encryptor.dms.endpoints import SourceEndpoint, TargetEndpoint
from rds_encryptor.dms.enums import MigrationType, ReplicationTaskStatus
from rds_encryptor.dms.replication_instance import ReplicationInstance
from rds_encryptor.utils import get_logger, normalize_aws_id


class TaskFailedException(Exception):
    def __init__(self, task, status, stop_reason, last_failure_message):
        self.task = task
        self.status = status
        self.stop_reason = stop_reason
        self.last_failure_message = last_failure_message


class TableMapping(NamedTuple):
    schema: Literal["%"] | str
    table: Literal["%"] | str
    action: Literal["include", "exclude"]


DEFAULT_REPLICATE_TASK_SETTINGS = {
    "Logging": {
        "EnableLogging": True,
        "EnableLogContext": True,
        "LogComponents": [
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "TRANSFORMATION"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "SOURCE_UNLOAD"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "IO"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "TARGET_LOAD"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "PERFORMANCE"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "SOURCE_CAPTURE"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "SORTER"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "REST_SERVER"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "VALIDATOR_EXT"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "TARGET_APPLY"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "TASK_MANAGER"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "TABLES_MANAGER"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "METADATA_MANAGER"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "FILE_FACTORY"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "COMMON"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "ADDONS"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "DATA_STRUCTURE"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "COMMUNICATION"},
            {"Severity": "LOGGER_SEVERITY_DEFAULT", "Id": "FILE_TRANSFER"},
        ],
    },
    "StreamBufferSettings": {
        "StreamBufferCount": 3,
        "CtrlStreamBufferSizeInMB": 5,
        "StreamBufferSizeInMB": 8,
    },
    "ErrorBehavior": {
        "FailOnNoTablesCaptured": True,
        "ApplyErrorUpdatePolicy": "LOG_ERROR",
        "FailOnTransactionConsistencyBreached": False,
        "RecoverableErrorThrottlingMax": 1800,
        "DataErrorEscalationPolicy": "SUSPEND_TABLE",
        "ApplyErrorEscalationCount": 0,
        "RecoverableErrorStopRetryAfterThrottlingMax": True,
        "RecoverableErrorThrottling": True,
        "ApplyErrorFailOnTruncationDdl": False,
        "DataMaskingErrorPolicy": "STOP_TASK",
        "DataTruncationErrorPolicy": "LOG_ERROR",
        "ApplyErrorInsertPolicy": "LOG_ERROR",
        "EventErrorPolicy": "IGNORE",
        "ApplyErrorEscalationPolicy": "LOG_ERROR",
        "RecoverableErrorCount": -1,
        "DataErrorEscalationCount": 0,
        "TableErrorEscalationPolicy": "STOP_TASK",
        "RecoverableErrorInterval": 5,
        "ApplyErrorDeletePolicy": "IGNORE_RECORD",
        "TableErrorEscalationCount": 0,
        "FullLoadIgnoreConflicts": True,
        "DataErrorPolicy": "LOG_ERROR",
        "TableErrorPolicy": "SUSPEND_TABLE",
    },
    "TTSettings": {"TTS3Settings": None, "TTRecordSettings": None, "EnableTT": False},
    "FullLoadSettings": {
        "CommitRate": 10000,
        "StopTaskCachedChangesApplied": False,
        "StopTaskCachedChangesNotApplied": False,
        "MaxFullLoadSubTasks": 8,
        "TransactionConsistencyTimeout": 600,
        "CreatePkAfterFullLoad": False,
        "TargetTablePrepMode": "DO_NOTHING",
    },
    "ValidationSettings": {
        "ValidationPartialLobSize": 0,
        "PartitionSize": 10000,
        "RecordFailureDelayLimitInMinutes": 0,
        "SkipLobColumns": False,
        "ValidationS3Time": 0,
        "FailureMaxCount": 10000,
        "HandleCollationDiff": False,
        "ValidationQueryCdcDelaySeconds": 0,
        "ValidationMode": "ROW_LEVEL",
        "TableFailureMaxCount": 1000,
        "RecordFailureDelayInMinutes": 5,
        "MaxKeyColumnSize": 8096,
        "EnableValidation": True,
        "ThreadCount": 5,
        "RecordSuspendDelayInMinutes": 30,
        "ValidationS3Mask": 0,
        "ValidationOnly": False,
    },
    "TargetMetadata": {
        "ParallelApplyBufferSize": 0,
        "ParallelApplyQueuesPerThread": 0,
        "ParallelApplyThreads": 0,
        "TargetSchema": "",
        "InlineLobMaxSize": 0,
        "ParallelLoadQueuesPerThread": 0,
        "SupportLobs": True,
        "LobChunkSize": 0,
        "TaskRecoveryTableEnabled": False,
        "ParallelLoadThreads": 0,
        "LobMaxSize": 100_000,  # TODO: Might be issue if your LOB e.g. JSONB field is larger than 100MB
        "BatchApplyEnabled": False,
        "FullLobMode": False,
        "LimitedSizeLobMode": True,
        "LoadMaxFileSize": 0,
        "ParallelLoadBufferSize": 0,
    },
    "BeforeImageSettings": None,
    "ControlTablesSettings": {
        "historyTimeslotInMinutes": 5,
        "HistoryTimeslotInMinutes": 5,
        "StatusTableEnabled": False,
        "SuspendedTablesTableEnabled": False,
        "HistoryTableEnabled": False,
        "ControlSchema": "",
        "FullLoadExceptionTableEnabled": False,
    },
    "LoopbackPreventionSettings": None,
    "CharacterSetSettings": None,
    "FailTaskWhenCleanTaskResourceFailed": False,
    "ChangeProcessingTuning": {
        "StatementCacheSize": 50,
        "CommitTimeout": 1,
        "RecoveryTimeout": -1,
        "BatchApplyPreserveTransaction": True,
        "BatchApplyTimeoutMin": 1,
        "BatchSplitSize": 0,
        "BatchApplyTimeoutMax": 30,
        "MinTransactionSize": 1000,
        "MemoryKeepTime": 60,
        "BatchApplyMemoryLimit": 500,
        "MemoryLimitTotal": 1024,
    },
    "ChangeProcessingDdlHandlingPolicy": {
        "HandleSourceTableDropped": True,
        "HandleSourceTableTruncated": True,
        "HandleSourceTableAltered": True,
    },
    "PostProcessingRules": None,
}

DEFAULT_REPLICATE_TASK_SETTINGS_JSON = json.dumps(DEFAULT_REPLICATE_TASK_SETTINGS)


class MigrationTask:
    logger = get_logger("MigrationTask")
    aws_client = boto3.client("dms")

    def __init__(self, task_id: str, arn: str):
        self.task_id = task_id
        self.arn = arn

    def _describe(self) -> dict:
        response = self.aws_client.describe_replication_tasks(
            Filters=[{"Name": "replication-task-id", "Values": [self.task_id]}]
        )
        if len(response["ReplicationTasks"]) == 0:
            raise ValueError(f"Replication task not found: {self.task_id}")
        if len(response["ReplicationTasks"]) > 1:
            raise ValueError(f"Multiple replication tasks found: {self.task_id}")
        return response["ReplicationTasks"][0]

    def get_status(self) -> ReplicationTaskStatus:
        return ReplicationTaskStatus(self._describe()["Status"])

    def _wait_until(
        self,
        expected_status: ReplicationTaskStatus,
        timeout: int = 4 * 60 * 60,
        pooling_frequency: int = 60,
    ) -> "MigrationTask":
        """
        :param expected_status: ReplicationTaskStatus
        :param timeout: Timeout in seconds. Default is 4 hours
        :param pooling_frequency: Pooling frequency in seconds. Default is 1 minute.
        :return: MigrationTask or raise TimeoutError
        """
        timeout_dt = datetime.now(tz=UTC) + timedelta(seconds=timeout)

        while datetime.now(tz=UTC) < timeout_dt:
            status = self.get_status()
            if status == expected_status:
                self.logger.info("Task %s is in status %s", self.task_id, status)
                return self
            self.logger.debug("Task %s is in status %s, waiting...", self.task_id, status)
            time.sleep(pooling_frequency)

        raise TimeoutError(f"Task {self.task_id} status is not {expected_status} after {timeout} seconds")

    def wait_until_ready(self) -> "MigrationTask":
        self.logger.info("Waiting for task %s to be ready ...", self.task_id)
        return self._wait_until(ReplicationTaskStatus.READY)

    def run_task(self) -> "MigrationTask":
        self.logger.info("Starting task %s ...", self.task_id)
        self.aws_client.start_replication_task(
            ReplicationTaskArn=self.arn, StartReplicationTaskType="start-replication"
        )
        self.logger.info("Task %s started", self.task_id)
        return self

    def wait_until_finished(self, timeout: int = 4 * 60 * 60, pooling_frequency: int = 2 * 60) -> "MigrationTask":
        timeout_dt = datetime.now(tz=UTC) + timedelta(seconds=timeout)

        while datetime.now(tz=UTC) < timeout_dt:
            response = self._describe()
            status = ReplicationTaskStatus(response["Status"])
            stop_reason = response.get("StopReason")
            last_failure_message = response.get("LastFailureMessage")
            full_load_progress = response.get("ReplicationTaskStats", {}).get("FullLoadProgressPercent", 0)
            if status == ReplicationTaskStatus.STOPPED and stop_reason == "Stop Reason NORMAL":
                self.logger.info("[Task %s] Task finished", self.task_id)
                return self
            if status == ReplicationTaskStatus.RUNNING and full_load_progress == 100:
                self.logger.info("[Task %s] Full load completed", self.task_id)
                return self
            if status in (ReplicationTaskStatus.STOPPED, ReplicationTaskStatus.FAILED):
                raise TaskFailedException(
                    task=self,
                    status=status,
                    stop_reason=stop_reason,
                    last_failure_message=last_failure_message,
                )

            self.logger.info(
                "[Task %s] Task status: %s, full load progress %s/100 waiting...",
                self.task_id,
                status,
                full_load_progress,
            )
            time.sleep(pooling_frequency)

        raise TimeoutError(f"Task {self.task_id} status is not STOPPED after {timeout} seconds")

    @classmethod
    def create_migration_task(
        cls,
        name,
        source_endpoint: SourceEndpoint,
        target_endpoint: TargetEndpoint,
        replication_instance: ReplicationInstance,
        migration_type: MigrationType,
        table_mappings: list[TableMapping],
        tags: list[dict[str, str]] = None,  # noqa: RUF013
    ):
        # TODO: Add check if the task already exists
        table_mappings_rules = {
            "rules": [
                {
                    "rule-type": "selection",
                    "rule-id": str(idx + 1),
                    "rule-name": str(idx + 1),
                    "object-locator": {
                        "schema-name": rule.schema,
                        "table-name": rule.table,
                    },
                    "rule-action": rule.action,
                    "filters": [],
                }
                for idx, rule in enumerate(table_mappings)
            ]
        }
        normalized_id = normalize_aws_id(name)
        cls.logger.info(
            'Creating migration task "%s" from "%s" to "%s" '
            'with migration type "%s" and table mappings "%s" on replication instance "%s" ...',
            normalized_id,
            source_endpoint.endpoint_id,
            target_endpoint.endpoint_id,
            str(migration_type),
            str(table_mappings),
            replication_instance.arn,
        )
        response = cls.aws_client.create_replication_task(
            ReplicationTaskIdentifier=normalized_id,
            SourceEndpointArn=source_endpoint.arn,
            TargetEndpointArn=target_endpoint.arn,
            ReplicationInstanceArn=replication_instance.arn,
            MigrationType=str(migration_type),
            TableMappings=json.dumps(table_mappings_rules),
            ReplicationTaskSettings=DEFAULT_REPLICATE_TASK_SETTINGS_JSON,
            Tags=tags or [],
        )["ReplicationTask"]
        # Replication Task is modifying the replication instance, so we need to wait until it's active
        replication_instance.wait_until_active()
        cls.logger.info('Migration task "%s" created', normalized_id)
        return cls(task_id=response["ReplicationTaskIdentifier"], arn=response["ReplicationTaskArn"])



import abc
import time
from datetime import UTC, datetime, timedelta
from typing import Literal, Optional

import boto3
from botocore.exceptions import ClientError

from rds_encryptor.rds.instance import RDSInstance
from rds_encryptor.utils import MIGRATION_SEED, get_logger, normalize_aws_id


class BaseEndpoint(abc.ABC):
    logger = get_logger("BaseEndpoint")
    aws_client = boto3.client("dms")
    endpoint_type: Literal["source", "target"]
    additional_settings: dict[str, str] = None

    def __init__(self, rds_instance: RDSInstance, database: str, kms_key_arn: str):
        self.rds_instance = rds_instance
        self.database = database
        self.endpoint_id = normalize_aws_id(
            f"{self.endpoint_type}-{self.rds_instance.instance_id}-{self.database}-{MIGRATION_SEED}"
        )
        self.kms_key_arn = kms_key_arn
        self._arn = None

    @property
    def arn(self) -> str:
        if self._arn is None:
            raise ValueError("You must call .get_or_create_endpoint() first to get the ARN")
        return self._arn

    def create_endpoint(self) -> "BaseEndpoint":
        assert self.endpoint_type in (
            "source",
            "target",
        ), f"Invalid endpoint type: {self.endpoint_type}"
        response = self.aws_client.create_endpoint(
            EndpointIdentifier=self.endpoint_id,
            EndpointType=self.endpoint_type,
            EngineName="postgres",
            KmsKeyId=self.kms_key_arn,
            PostgreSQLSettings={
                "DatabaseName": self.database,
                "Port": self.rds_instance.port,
                "ServerName": self.rds_instance.endpoint,
                "Username": self.rds_instance.master_username,
                "Password": self.rds_instance.master_password,
                **(self.additional_settings or {}),
            },
            Tags=self.rds_instance.tags,
        )
        self.logger.info('%s endpoint "%s" created', self.endpoint_type.capitalize(), self.endpoint_id)
        self._arn = response["Endpoint"]["EndpointArn"]
        return self

    def _describe(self):
        response = self.aws_client.describe_endpoints(Filters=[{"Name": "endpoint-id", "Values": [self.endpoint_id]}])
        if len(response["Endpoints"]) > 1:
            raise ValueError(f"Multiple endpoints found: {self.endpoint_id}")

        return response["Endpoints"][0]

    def get_status(self) -> str:
        return self._describe()["Status"]

    def get_endpoint(self) -> Optional["BaseEndpoint"]:
        try:
            self._arn = self._describe()["EndpointArn"]
        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundFault":
                return None
            raise
        return self

    def get_or_create_endpoint(self) -> "BaseEndpoint":
        self.logger.info(
            'Creating %s endpoint "%s" for "%s" database', self.endpoint_type, self.endpoint_id, self.database
        )
        endpoint = self.get_endpoint()
        if self._arn is not None:
            self.logger.info('Endpoint "%s" already exists', self.endpoint_id)
            return endpoint
        return self.create_endpoint()

    def wait_until_created(self, timeout: int = 60 * 60, pooling_frequency: int = 30) -> "BaseEndpoint":
        timeout_dt = datetime.now(tz=UTC) + timedelta(seconds=timeout)
        self.logger.info('Waiting for %s endpoint "%s" to become available ...', self.endpoint_type, self.endpoint_id)

        while datetime.now(tz=UTC) < timeout_dt:
            status = self.get_status()
            if status == "active":
                self.logger.info('Endpoint "%s" is active', self.endpoint_id)
                return self
            self.logger.debug("Endpoint %s is in status %s, waiting...", self.endpoint_id, status)
            time.sleep(pooling_frequency)

        raise TimeoutError(f"Endpoint {self.endpoint_id} creation timeout")


class SourceEndpoint(BaseEndpoint):
    endpoint_type = "source"


class TargetEndpoint(BaseEndpoint):
    endpoint_type = "target"
    additional_settings = {
        "AfterConnectScript": "SET session_replication_role = replica",
    }



import abc
from collections.abc import Generator

import psycopg2

from rds_encryptor.rds.instance import RDSInstance


class InvalidCredentialsException(Exception):
    pass


class InvalidPostgresCredentialsException(InvalidCredentialsException):
    pass


class DBManager(abc.ABC):
    invalid_credentials_exception: InvalidCredentialsException

    @staticmethod
    def from_rds(rds_instance: RDSInstance, database: str = "postgres") -> "PostgresDBManager":
        return PostgresDBManager(
            host=rds_instance.endpoint,
            port=rds_instance.port,
            user=rds_instance.master_username,
            password=rds_instance.master_password,
            database=database,
        )

    @abc.abstractmethod
    def check_connection(self) -> bool:
        pass


class PostgresDBManager:
    invalid_credentials_exception = InvalidPostgresCredentialsException

    def __init__(self, host: str, port: int, user: str, password: str, database: str):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database

    def __get_connection(self) -> psycopg2.extensions.connection:
        return psycopg2.connect(
            host=self.host,
            port=self.port,
            user=self.user,
            password=self.password,
            database=self.database,
        )

    def check_connection(self) -> bool:
        try:
            conn = self.__get_connection()
            cursor = conn.cursor()
            cursor.execute("SELECT 1")
            cursor.close()
            conn.close()
        except psycopg2.DatabaseError:
            return False
        return True

    def get_parameter(self, parameter: str) -> str:
        conn = self.__get_connection()
        cursor = conn.cursor()
        cursor.execute(f"SHOW {parameter}")
        result = cursor.fetchone()[0]
        cursor.close()
        conn.close()
        return result

    def create_extension(self, extension: str):
        conn = self.__get_connection()
        cursor = conn.cursor()
        cursor.execute(f"CREATE EXTENSION IF NOT EXISTS {extension}")
        conn.commit()
        cursor.close()
        conn.close()

    def get_all_tables(self) -> list[str]:
        conn = self.__get_connection()
        cursor = conn.cursor()
        cursor.execute(
            """
            SELECT schemaname, tablename
            FROM pg_catalog.pg_tables
            WHERE schemaname NOT LIKE 'pg_%'
              AND schemaname != 'information_schema'
              and tablename not like 'awsdms_ddl_audit%'
            ORDER BY schemaname, tablename;
            """
        )
        tables = [f"{row[0]}.{row[1]}" for row in cursor.fetchall()]
        cursor.close()
        conn.close()
        return tables

    def truncate_database(self):
        conn = self.__get_connection()
        cursor = conn.cursor()
        cursor.execute(
            """
            SELECT schema_name
            FROM information_schema.schemata
            WHERE (schema_name NOT LIKE 'pg_%'
              AND schema_name != 'information_schema') or schema_name = 'pglogical';
        """
        )
        schemas = [row[0] for row in cursor.fetchall()]
        for schema in schemas:
            # FIXME: S608 Possible SQL injection vector through string-based query construction
            cursor.execute(f"SELECT tablename FROM pg_tables WHERE schemaname = '{schema}'")  # noqa: S608
            tables = [row[0] for row in cursor.fetchall()]
            for table in tables:
                cursor.execute(f"TRUNCATE TABLE {schema}.{table} CASCADE")
        conn.commit()
        cursor.close()
        conn.close()

    def get_sequences(self) -> list[dict[str, int | str]]:
        conn = self.__get_connection()
        cursor = conn.cursor()
        cursor.execute("SELECT schemaname, sequencename, last_value FROM pg_sequences;")
        sequences = [
            {
                "schema": row[0],
                "sequence": row[1],
                "last_value": row[2] or 1,
            }
            for row in cursor.fetchall()
        ]
        cursor.close()
        conn.close()
        return sequences

    def set_sequences(self, sequences: list[dict[str, int | str]]):
        conn = self.__get_connection()
        cursor = conn.cursor()
        for sequence in sequences:
            if sequence["sequence"].startswith("awsdms_ddl_audit"):
                continue
            cursor.execute(f"SELECT setval('{sequence['schema']}.{sequence['sequence']}', {sequence['last_value']})")
        conn.commit()
        cursor.close()
        conn.close()

    def iter_count(self, tables: list[str]) -> Generator[int, None, None]:
        conn = self.__get_connection()
        cursor = conn.cursor()
        for table in tables:
            cursor.execute(f"SELECT COUNT(*) FROM {table}")  # noqa: S608
            yield cursor.fetchone()[0]
        cursor.close()
        conn.close()



import time
from datetime import UTC, datetime, timedelta
from typing import TYPE_CHECKING, Optional

import boto3
from botocore.exceptions import ClientError

from rds_encryptor.utils import get_logger

if TYPE_CHECKING:
    from rds_encryptor.rds.instance import RDSInstance


class RDSSnapshot:
    logger = get_logger("RDSSnapshot")
    aws_client = boto3.client("rds")

    def __init__(self, snapshot_id: str, arn: str, tags: list[dict] = None):  # noqa: RUF013
        self.snapshot_id = snapshot_id
        self.arn = arn
        if tags is None:
            tags = []
        self.tags = tags

    def get_status(self):
        snapshot = self.aws_client.describe_db_snapshots(
            DBSnapshotIdentifier=self.snapshot_id,
        )["DBSnapshots"][0]
        return snapshot["Status"]

    @classmethod
    def from_id(cls, snapshot_id: str) -> Optional["RDSSnapshot"]:
        assert snapshot_id, "Snapshot ID is required"

        try:
            snapshots = cls.aws_client.describe_db_snapshots(
                DBSnapshotIdentifier=snapshot_id,
            )["DBSnapshots"]
        except ClientError as e:
            if e.response["Error"]["Code"] == "DBSnapshotNotFound":
                return None
            raise
        if len(snapshots) > 1:
            raise ValueError(f"Multiple snapshots found: {snapshot_id}")

        snapshot = snapshots[0]

        return cls(
            snapshot_id=snapshot_id,
            arn=snapshot["DBSnapshotArn"],
            tags=snapshot.get("TagList"),
        )

    def copy_snapshot(
        self,
        encryption_kms_key_arn: str,
        copy_tags: bool = True,
    ) -> "RDSSnapshot":
        target_snapshot_id = f"{self.snapshot_id}-encrypted"
        self.logger.info('Copying and encrypting snapshot "%s" to "%s" ...', self.snapshot_id, target_snapshot_id)
        target_snapshot = self.from_id(snapshot_id=target_snapshot_id)
        if target_snapshot is not None:
            self.logger.info('Snapshot "%s" already exists', target_snapshot_id)
            return target_snapshot

        response = self.aws_client.copy_db_snapshot(
            SourceDBSnapshotIdentifier=self.arn,
            TargetDBSnapshotIdentifier=target_snapshot_id,
            SourceRegion=self.aws_client.meta.region_name,
            KmsKeyId=encryption_kms_key_arn,
            CopyTags=copy_tags,
            Tags=self.tags,
        )
        self.logger.info('Snapshot "%s" is being copied', target_snapshot_id)
        return RDSSnapshot.from_id(response["DBSnapshot"]["DBSnapshotIdentifier"])

    def wait_until_created(self, timeout: int = 60 * 60, pooling_frequency: int = 60) -> "RDSSnapshot":
        self.logger.info('Waiting for snapshot "%s" to become available ...', self.snapshot_id)
        timeout_dt = datetime.now(tz=UTC) + timedelta(seconds=timeout)

        while datetime.now(tz=UTC) < timeout_dt:
            status = self.get_status()
            if status == "available":
                self.logger.info('Snapshot "%s" is available', self.snapshot_id)
                return self
            if status == "failed":
                raise ValueError(f"Snapshot {self.snapshot_id} creation failed")
            time.sleep(pooling_frequency)

        raise TimeoutError(f"Snapshot {self.snapshot_id} creation timeout after {timeout} seconds")

    def restore_snapshot(
        self,
        instance_identifier: str,
        from_rds_instance: "RDSInstance",
        master_password: str,
        tags: list[dict[str, str]] = None,  # noqa: RUF013
    ) -> "RDSInstance":
        from rds_encryptor.rds.instance import RDSInstance

        old_instance = from_rds_instance._describe()

        tags = tags or []
        self.logger.info('Restoring snapshot "%s" to instance "%s" ...', self.snapshot_id, instance_identifier)
        response = self.aws_client.restore_db_instance_from_db_snapshot(
            DBInstanceIdentifier=instance_identifier,
            DBSnapshotIdentifier=self.snapshot_id,
            PubliclyAccessible=old_instance["PubliclyAccessible"],
            DBSubnetGroupName=old_instance["DBSubnetGroup"]["DBSubnetGroupName"],
            VpcSecurityGroupIds=[sg["VpcSecurityGroupId"] for sg in old_instance["VpcSecurityGroups"]],
            CopyTagsToSnapshot=old_instance["CopyTagsToSnapshot"],
            Tags=tags,
            AvailabilityZone=old_instance["AvailabilityZone"],
        )
        self.logger.info('Instance "%s" is being restored', instance_identifier)
        return RDSInstance.from_id(
            instance_id=response["DBInstance"]["DBInstanceIdentifier"], root_password=master_password
        )



from enum import StrEnum


class MigrationType(StrEnum):
    migrate = "full-load"
    replicate = "cdc"
    migrate_replicate = "full-load-and-cdc"


class ReplicationTaskStatus(StrEnum):
    """
    "moving" - The task is being moved in response to running the MoveReplicationTask operation.
    "creating" - The task is being created in response to running the CreateReplicationTask operation.
    "deleting" - The task is being deleted in response to running the DeleteReplicationTask operation.
    "failed" - The task failed to successfully complete the database migration in response to running the StartReplicationTask operation.
    "failed-move" - The task failed to move in response to running the MoveReplicationTask operation.
    "modifying" - The task definition is being modified in response to running the ModifyReplicationTask operation.
    "ready" - The task is in a ready state where it can respond to other task operations, such as StartReplicationTask or DeleteReplicationTask.
    "running" - The task is performing a database migration in response to running the StartReplicationTask operation.
    "starting" - The task is preparing to perform a database migration in response to running the StartReplicationTask operation.
    "stopped" - The task has stopped in response to running the StopReplicationTask operation.
    "stopping" - The task is preparing to stop in response to running the StopReplicationTask operation.
    "testing" - The database migration specified for this task is being tested in response to running either the StartReplicationTaskAssessmentRun or the StartReplicationTaskAssessment operation.
    """  # noqa: E501

    MOVING = "moving"
    CREATING = "creating"
    DELETING = "deleting"
    FAILED = "failed"
    FAILED_MOVE = "failed-move"
    MODIFYING = "modifying"
    READY = "ready"
    RUNNING = "running"
    STARTING = "starting"
    STOPPED = "stopped"
    STOPPING = "stopping"
    TESTING = "testing"



import argparse

from rds_encryptor.encryption_pipeline import EncryptionPipeline


def main():
    parser = argparse.ArgumentParser(description="Encrypt RDS instance")
    parser.add_argument("--rds-instance-name", "-r", type=str, required=True, help="RDS instance ID")
    parser.add_argument(
        "--master-password",
        "-p",
        type=str,
        required=True,
        help="Master password for the database",
    )
    parser.add_argument(
        "--kms-key-arn",
        "-k",
        type=str,
        required=True,
        help="AWS KMS key ARN for encryption",
    )
    parser.add_argument(
        "--dms-replication-instance-arn",
        "-i",
        type=str,
        required=True,
        help="AWS DMS replication instance ARN",
    )
    parser.add_argument(
        "--databases",
        "-d",
        type=str,
        nargs="*",
        required=True,
        help="List of databases to encrypt",
    )
    parser.add_argument(
        "--new-instance-identifier",
        "-n",
        type=str,
        required=False,
        help="Identifier for the new encrypted RDS instance",
    )
    args = parser.parse_args()
    pipeline = EncryptionPipeline(
        instance_id=args.rds_instance_name,
        master_password=args.master_password,
        kms_key_arn=args.kms_key_arn,
        dms_replication_instance_arn=args.dms_replication_instance_arn,
        databases=args.databases,
        new_instance_identifier=args.new_instance_identifier,
    )
    pipeline.run_pipeline()


if __name__ == "__main__":
    main()



from typing import Optional

import boto3
from botocore.exceptions import ClientError

from rds_encryptor.utils import MIGRATION_SEED, get_logger


def build_shared_preload_libraries_param(*libraries: str) -> str:
    return ",".join(map(str.strip, libraries))


def get_migration_parameter_group_name(parameter_group_name: str) -> str:
    salt = f"-{MIGRATION_SEED}-migration"
    if parameter_group_name.endswith(salt):
        return parameter_group_name
    return f"{parameter_group_name}{salt}"


def get_original_parameter_group(migration_parameter_group_name) -> str:
    return migration_parameter_group_name.replace(f"-{MIGRATION_SEED}-migration", "")


class ParameterGroup:
    aws_client = boto3.client("rds")
    logger = get_logger("ParameterGroup")

    def __init__(self, name: str):
        self.name = name
        self.properties = self._fetch_properties()

    @classmethod
    def from_name(cls, name: str) -> Optional["ParameterGroup"]:
        assert name, "Parameter group name is required"

        try:
            response = cls.aws_client.describe_db_parameter_groups(
                DBParameterGroupName=name,
            )["DBParameterGroups"]
        except ClientError as e:
            if e.response["Error"]["Code"] == "DBParameterGroupNotFound":
                return None
            raise
        if len(response) > 1:
            raise ValueError(f"Multiple parameter groups found: {name}")

        return cls(name=name)

    def copy(self) -> "ParameterGroup":
        new_parameter_group_name = get_migration_parameter_group_name(self.name)
        self.logger.info('Copying parameter group "%s" to "%s" ...', self.name, new_parameter_group_name)
        response = self.aws_client.copy_db_parameter_group(
            SourceDBParameterGroupIdentifier=self.name,
            TargetDBParameterGroupIdentifier=new_parameter_group_name,
            TargetDBParameterGroupDescription=f"{self.name} migration parameter group",
        )
        self.logger.info('Parameter group "%s" copied', new_parameter_group_name)
        return ParameterGroup(name=response["DBParameterGroup"]["DBParameterGroupName"])

    def _fetch_properties(self):
        response = self.aws_client.describe_db_parameters(
            DBParameterGroupName=self.name,
        )
        parameters = response.get("Parameters", [])
        while response.get("Marker"):
            response = self.aws_client.describe_db_parameters(
                DBParameterGroupName=self.name,
                Marker=response["Marker"],
            )
            parameters.extend(response.get("Parameters", []))
        return {
            param["ParameterName"]: {"value": param["ParameterValue"], "apply_type": param["ApplyType"]}
            for param in parameters
            if "ParameterValue" in param
        }

    @property
    def wal_sender_timeout(self) -> int:
        return int(self.properties.get("wal_sender_timeout", {"value": 0})["value"])

    @property
    def shared_preload_libraries(self) -> list[str]:
        return list(
            map(
                str.strip,
                self.properties.get("shared_preload_libraries", {"value": ""})["value"].split(","),
            )
        )

    @property
    def rds_logical_replication(self) -> int:
        return int(self.properties.get("rds.logical_replication", {"value": 0})["value"])

    def set_parameter(self, name: str, value: any) -> None:
        parameter = {"ParameterName": name, "ParameterValue": str(value)}
        old_parameter = self.properties.get(name)
        if old_parameter is not None and old_parameter["apply_type"] != "static":
            parameter["ApplyMethod"] = "immediate"
        else:
            parameter["ApplyMethod"] = "pending-reboot"
        self.aws_client.modify_db_parameter_group(
            DBParameterGroupName=self.name,
            Parameters=[parameter],
        )
        self.properties = self._fetch_properties()

    def delete(self) -> None:
        self.logger.info('Deleting parameter group "%s" ...', self.name)
        self.aws_client.delete_db_parameter_group(
            DBParameterGroupName=self.name,
        )
        self.logger.info('Parameter group "%s" deleted', self.name)



import logging
import os
import sys
import time
from datetime import UTC, datetime

DEBUG = bool(int(os.getenv("DEBUG", "0")))
MIGRATION_SEED = datetime.now(tz=UTC).strftime("%Y-%m-%d")


def get_logger(name: str = "deploy", log_level: str = logging.INFO) -> logging.Logger:
    if DEBUG:
        log_level = logging.DEBUG
    logger = logging.getLogger(name)
    stream_handler = logging.StreamHandler(sys.stdout)
    log_formatter = logging.Formatter(
        "[%(name)s] [%(levelname)s] %(asctime)s: %(message)s",
    )
    log_formatter.converter = time.gmtime
    stream_handler.setFormatter(log_formatter)
    logger.addHandler(stream_handler)
    logger.setLevel(log_level)
    return logger


def normalize_aws_id(db_name: str) -> str:
    """
    Identifiers must begin with a letter and must contain only ASCII letters, digits, and hyphens.
    They can't end with a hyphen, or contain two consecutive hyphens.
    """
    return "".join(c for c in db_name.replace("_", "-") if c.isalnum() or c == "-").strip("-")



import time
from datetime import UTC, datetime, timedelta
from typing import Optional

import boto3
from botocore.exceptions import ClientError

from rds_encryptor.rds.parameter_group import ParameterGroup
from rds_encryptor.rds.snapshot import RDSSnapshot
from rds_encryptor.utils import MIGRATION_SEED, get_logger


class RDSInstance:
    logger = get_logger("RDSInstance")
    aws_client = boto3.client("rds")

    def __init__(
        self,
        instance_id: str,
        endpoint: str | None,
        port: int | None,
        master_username: str,
        master_password: str,
        parameter_group: ParameterGroup,
        tags: list[dict[str, str]] = None,  # noqa: RUF013
    ):
        self.instance_id = instance_id
        self._endpoint = endpoint
        self._port = port
        self.master_username = master_username
        self.master_password = master_password
        self.parameter_group = parameter_group
        if tags is None:
            tags = []
        self.tags = tags

    @property
    def endpoint(self) -> str:
        if self._endpoint is None:
            raise ValueError("You must call .wait_until_available() first to get the endpoint")
        return self._endpoint

    @property
    def port(self) -> int:
        if self._port is None:
            raise ValueError("You must call .wait_until_available() first to get the port")
        return self._port

    def _describe(self) -> dict:
        instances = self.aws_client.describe_db_instances(
            DBInstanceIdentifier=self.instance_id,
        )["DBInstances"]
        if len(instances) > 1:
            raise ValueError(f"Multiple instances found: {self.instance_id}")
        return instances[0]

    @classmethod
    def from_id(cls, instance_id: str, root_password: str) -> Optional["RDSInstance"]:
        assert instance_id, "Instance ID is required"
        assert root_password, "Root password is required"

        try:
            instances = cls.aws_client.describe_db_instances(
                DBInstanceIdentifier=instance_id,
            )["DBInstances"]
        except ClientError as e:
            if e.response["Error"]["Code"] == "DBInstanceNotFound":
                return None
            raise
        if len(instances) > 1:
            raise ValueError(f"Multiple instances found: {instance_id}")

        instance = instances[0]
        is_creating = instance["DBInstanceStatus"] == "creating"

        return cls(
            instance_id=instance_id,
            endpoint=None if is_creating else instance["Endpoint"]["Address"],
            port=None if is_creating else instance["Endpoint"]["Port"],
            master_username=instance["MasterUsername"],
            master_password=root_password,
            parameter_group=ParameterGroup.from_name(instance["DBParameterGroups"][0]["DBParameterGroupName"]),
            tags=instance.get("TagList"),
        )

    def get_status(self) -> str:
        instance = self._describe()
        return instance["DBInstanceStatus"]

    def take_snapshot(self) -> RDSSnapshot:
        snapshot_id = f"{self.instance_id}-{MIGRATION_SEED}-migration"
        self.logger.info('Taking snapshot "%s" for instance "%s" ...', snapshot_id, self.instance_id)
        snapshot = RDSSnapshot.from_id(snapshot_id)
        if snapshot is not None:
            self.logger.info('Snapshot "%s" already exists, skipping...', snapshot_id)
            return snapshot

        response = self.aws_client.create_db_snapshot(
            DBSnapshotIdentifier=snapshot_id,
            DBInstanceIdentifier=self.instance_id,
            Tags=self.tags,
        )["DBSnapshot"]
        self.logger.info('Snapshot "%s" created', snapshot_id)
        return RDSSnapshot.from_id(snapshot_id=response["DBSnapshotIdentifier"])

    def set_parameter_group(self, parameter_group: ParameterGroup) -> "RDSInstance":
        self.logger.info(
            'Setting "%s" parameter group for "%s" instance...',
            self.instance_id,
            parameter_group.name,
        )
        self.aws_client.modify_db_instance(
            DBInstanceIdentifier=self.instance_id,
            DBParameterGroupName=parameter_group.name,
            ApplyImmediately=True,
        )
        self.parameter_group = parameter_group
        time.sleep(20)  # Wait for the parameter group to be applied
        self.logger.info(
            'Parameter group "%s" set for "%s" instance',
            parameter_group.name,
            self.instance_id,
        )
        return self

    def modify_instance(
        self,
        **params,
    ) -> "RDSInstance":
        params_str = ", ".join(f"{k}={v}" for k, v in params.items())
        self.logger.info('Modifying "%s" instance with %s ...', self.instance_id, params_str)
        self.aws_client.modify_db_instance(
            DBInstanceIdentifier=self.instance_id,
            ApplyImmediately=True,
            **params,
        )
        self.logger.info('"%s" instance modified', self.instance_id)
        return self

    def wait_until_available(self, timeout: int = 60 * 60, pooling_frequency: int = 30) -> "RDSInstance":
        timeout_dt = datetime.now(tz=UTC) + timedelta(seconds=timeout)
        self.logger.info('Waiting for instance "%s" to become available ...', self.instance_id)

        while datetime.now(tz=UTC) < timeout_dt:
            instance = self._describe()
            status = instance["DBInstanceStatus"]
            if status == "available":
                self.logger.info('Instance "%s" is available', self.instance_id)
                self._endpoint = instance["Endpoint"]["Address"]
                self._port = instance["Endpoint"]["Port"]
                return self
            time.sleep(pooling_frequency)

        raise TimeoutError(f"Instance {self.instance_id} is not available after {timeout} seconds")



from rds_encryptor.db_manager import DBManager
from rds_encryptor.dms.endpoints import SourceEndpoint, TargetEndpoint
from rds_encryptor.dms.enums import MigrationType
from rds_encryptor.dms.migration_task import MigrationTask, TableMapping
from rds_encryptor.dms.replication_instance import ReplicationInstance
from rds_encryptor.dms.task_manager import MigrationTaskManager
from rds_encryptor.rds.instance import RDSInstance
from rds_encryptor.rds.parameter_group import (
    ParameterGroup,
    build_shared_preload_libraries_param,
    get_migration_parameter_group_name,
    get_original_parameter_group,
)
from rds_encryptor.utils import MIGRATION_SEED, get_logger, normalize_aws_id


class EncryptionPipeline:
    logger = get_logger("EncryptionPipeline")

    def __init__(
        self,
        instance_id: str,
        master_password: str,
        kms_key_arn: str,
        dms_replication_instance_arn: str,
        databases: list[str] = None,  # noqa: RUF013
        new_instance_identifier: str | None = None,
    ):
        self.rds_instance = RDSInstance.from_id(instance_id=instance_id, root_password=master_password)
        if self.rds_instance is None:
            raise ValueError(f"Cannot find source RDS instance by identifier={instance_id}")

        self.kms_key_arn = kms_key_arn
        self.dms_replication_instance_arn = dms_replication_instance_arn
        self.databases = databases or []
        self.new_instance_identifier = new_instance_identifier or normalize_aws_id(
            f"{instance_id}-{MIGRATION_SEED}-encrypted"
        )

    def check_databases_connections(self):
        self.logger.info("Checking database connections...")
        for database in self.databases:
            db_manager = DBManager.from_rds(rds_instance=self.rds_instance, database=database)
            if not db_manager.check_connection():
                raise db_manager.invalid_credentials_exception(f"Cannot connect to source RDS to {database} database.")
            self.logger.info('Successfully connected to "%s" database.', database)

    def create_encrypted_instance(self):
        self.logger.info('Trying to provision encrypted RDS instance with ID: "%s" ...', self.new_instance_identifier)
        encrypted_rds_instance: RDSInstance | None = RDSInstance.from_id(
            instance_id=self.new_instance_identifier,
            root_password=self.rds_instance.master_password,
        )
        if encrypted_rds_instance is None:
            snapshot = self.rds_instance.take_snapshot().wait_until_created()
            encrypted_snapshot = snapshot.copy_snapshot(
                copy_tags=True,
                encryption_kms_key_arn=self.kms_key_arn,
            ).wait_until_created()
            encrypted_rds_instance = encrypted_snapshot.restore_snapshot(
                instance_identifier=self.new_instance_identifier,
                from_rds_instance=self.rds_instance,
                master_password=self.rds_instance.master_password,
                tags=self.rds_instance.tags,
            ).wait_until_available()

            rds_instance_params = self.rds_instance._describe()
            encrypted_rds_instance.modify_instance(
                DBSecurityGroups=rds_instance_params["DBSecurityGroups"],
                DatabaseInsightsMode=rds_instance_params["DatabaseInsightsMode"],
                EnablePerformanceInsights=rds_instance_params["PerformanceInsightsEnabled"],
                PerformanceInsightsKMSKeyId=self.kms_key_arn,
                MaxAllocatedStorage=rds_instance_params["MaxAllocatedStorage"],
            ).wait_until_available()
        else:
            self.logger.info(
                'Skip provisioning "%s" instance, because it\'s already provisioned', self.new_instance_identifier
            )
        return encrypted_rds_instance

    def create_parameter_group_for_dms(self) -> ParameterGroup:
        parameter_group_name = get_migration_parameter_group_name(self.rds_instance.parameter_group.name)
        migration_parameter_group: ParameterGroup | None = ParameterGroup.from_name(name=parameter_group_name)
        self.logger.info('Creating parameter group "%s" for DMS migration...', parameter_group_name)

        if migration_parameter_group is None:
            migration_parameter_group: ParameterGroup = self.rds_instance.parameter_group.copy()
        if migration_parameter_group.wal_sender_timeout != 0:
            self.logger.info(
                "%s.wal_sender_timeout=%s, setting to 0",
                parameter_group_name,
                migration_parameter_group.wal_sender_timeout,
            )
            migration_parameter_group.set_parameter("wal_sender_timeout", 0)
        if "pglogical" not in migration_parameter_group.shared_preload_libraries:
            self.logger.info(
                "pglogical not in %s.shared_preload_libraries=%s, adding pglogical",
                parameter_group_name,
                migration_parameter_group.shared_preload_libraries,
            )
            migration_parameter_group.set_parameter(
                "shared_preload_libraries",
                build_shared_preload_libraries_param("pglogical", *migration_parameter_group.shared_preload_libraries),
            )
        if migration_parameter_group.rds_logical_replication != 1:
            self.logger.info(
                "%s.rds_logical_replication=%s, setting to 1",
                parameter_group_name,
                migration_parameter_group.rds_logical_replication,
            )
            migration_parameter_group.set_parameter("rds.logical_replication", 1)

        return migration_parameter_group

    def create_pglogical_extension_in_source_db(self):
        db_manager = DBManager.from_rds(rds_instance=self.rds_instance)
        while "pglogical" not in db_manager.get_parameter(
            "shared_preload_libraries"
        ):  # TODO: potential bug if any substring contains 'pglogical'
            self.logger.error(
                "`pglogical` not found in `shared_preload_libraries`, "
                'restart "%s" database to apply it and hit <Enter>.',
                self.rds_instance.instance_id,
            )
            input()

        for database in self.databases:
            source_db_manager = DBManager.from_rds(rds_instance=self.rds_instance, database=database)
            self.logger.info('Creating "pglogical" extension for %s database', database)
            source_db_manager.create_extension("pglogical")

    def create_replication_tasks(self, encrypted_rds_instance: RDSInstance) -> MigrationTaskManager:
        task_manager = MigrationTaskManager()
        dms_replication_instance = ReplicationInstance.from_arn(arn=self.dms_replication_instance_arn)

        for database in self.databases:
            source_endpoint = (
                SourceEndpoint(self.rds_instance, database=database, kms_key_arn=self.kms_key_arn)
                .get_or_create_endpoint()
                .wait_until_created()
            )
            target_endpoint = (
                TargetEndpoint(
                    encrypted_rds_instance,
                    database=database,
                    kms_key_arn=self.kms_key_arn,
                )
                .get_or_create_endpoint()
                .wait_until_created()
            )
            migration_task = MigrationTask.create_migration_task(
                name=normalize_aws_id(f"{self.rds_instance.instance_id}-{database}-{MIGRATION_SEED}"),
                source_endpoint=source_endpoint,
                target_endpoint=target_endpoint,
                replication_instance=dms_replication_instance,
                migration_type=MigrationType.migrate_replicate,
                table_mappings=[
                    TableMapping(schema="%", table="%", action="include"),
                    TableMapping(schema="pg_%", table="%", action="exclude"),
                    TableMapping(schema="information_schema", table="%", action="exclude"),
                    TableMapping(schema="pglogical", table="%", action="exclude"),
                ],
                tags=self.rds_instance.tags,
            )

            self.logger.info(
                'Truncating tables in "%s" database for instance "%s" ...', database, encrypted_rds_instance.instance_id
            )
            DBManager.from_rds(rds_instance=encrypted_rds_instance, database=database).truncate_database()
            self.logger.info(
                'Tables truncated in "%s" database for instance "%s"', database, encrypted_rds_instance.instance_id
            )
            task_manager.add_task(migration_task)

        return task_manager

    def migrate_databases_sequences(self, encrypted_rds_instance: RDSInstance):
        for database in self.databases:
            self.logger.info(
                'Start migrating "%s" database sequences from "%s" to "%s" instance...',
                database,
                self.rds_instance.instance_id,
                encrypted_rds_instance.instance_id,
            )
            source_db_manager = DBManager.from_rds(rds_instance=self.rds_instance, database=database)
            target_db_manager = DBManager.from_rds(rds_instance=encrypted_rds_instance, database=database)
            sequences = source_db_manager.get_sequences()
            target_db_manager.set_sequences(sequences)
            self.logger.info(
                'Sequences migrated for "%s" database from "%s" to "%s" instance',
                database,
                self.rds_instance.instance_id,
                encrypted_rds_instance.instance_id,
            )

    def check_data_consistency(self, encrypted_rds_instance: RDSInstance):
        for database in self.databases:
            source_db_manager = DBManager.from_rds(rds_instance=self.rds_instance, database=database)
            target_db_manager = DBManager.from_rds(rds_instance=encrypted_rds_instance, database=database)
            self.logger.info(
                'Checking data consistency for "%s" database between "%s" and "%s" instances...',
                database,
                self.rds_instance.instance_id,
                encrypted_rds_instance.instance_id,
            )
            tables = source_db_manager.get_all_tables()

            iterator = zip(
                tables, source_db_manager.iter_count(tables), target_db_manager.iter_count(tables), strict=True
            )

            diff_count = {}

            for table, source_count, target_count in iterator:
                if source_count != target_count:
                    diff_count[table] = (source_count, target_count)
                    self.logger.error(
                        'Data inconsistency for table "%s" between "%s" and "%s" instances: '
                        "source count=%s, target count=%s",
                        table,
                        self.rds_instance.instance_id,
                        encrypted_rds_instance.instance_id,
                        source_count,
                        target_count,
                    )

            if not diff_count:
                self.logger.info(
                    'Data consistency check for "%s" database between "%s" and "%s" instances passed',
                    database,
                    self.rds_instance.instance_id,
                    encrypted_rds_instance.instance_id,
                )
            else:
                self.logger.error(
                    'Data consistency check for "%s" database between "%s" and "%s" instances failed',
                    database,
                    self.rds_instance.instance_id,
                    encrypted_rds_instance.instance_id,
                )

    def rollback_parameter_group(self, encrypted_rds_instance: RDSInstance):
        # TODO: DEPRECATED
        original_parameter_group_name = get_original_parameter_group(self.rds_instance.parameter_group.name)
        parameter_group: ParameterGroup | None = ParameterGroup.from_name(name=original_parameter_group_name)
        if parameter_group is None:
            self.logger.warning('Cannot find original parameter group "%s" for rollback', original_parameter_group_name)
            return
        migration_parameter_group = self.rds_instance.parameter_group
        self.rds_instance.set_parameter_group(parameter_group).wait_until_available()
        encrypted_rds_instance.set_parameter_group(parameter_group).wait_until_available()
        input(f'Please reboot "{self.rds_instance.instance_id}" database and hit <Enter>')
        self.rds_instance.wait_until_available()
        input(f'Please reboot "{encrypted_rds_instance.instance_id}" database and hit <Enter>')
        encrypted_rds_instance.wait_until_available()
        migration_parameter_group.delete()
        self.logger.info('Rollback to "%s" parameter group finished', original_parameter_group_name)

    def run_pipeline(self):
        self.check_databases_connections()
        encrypted_rds_instance = self.create_encrypted_instance()
        migration_parameter_group = self.create_parameter_group_for_dms()

        # TODO: Need to set previous parameter group after migration
        if self.rds_instance.parameter_group.name != migration_parameter_group.name:
            self.rds_instance.wait_until_available().set_parameter_group(
                migration_parameter_group
            ).wait_until_available()
            input(f'Please reboot "{self.rds_instance.instance_id}" database and hit <Enter>')
            self.rds_instance.wait_until_available()
        if encrypted_rds_instance.parameter_group.name != migration_parameter_group.name:
            encrypted_rds_instance.wait_until_available().set_parameter_group(
                migration_parameter_group
            ).wait_until_available()
            input(f'Please reboot "{encrypted_rds_instance.instance_id}" database and hit <Enter>')
            encrypted_rds_instance.wait_until_available()
        self.create_pglogical_extension_in_source_db()

        task_manager = self.create_replication_tasks(encrypted_rds_instance)

        if task_manager.run_all():
            self.logger.info("All tasks finished successfully.")
            self.migrate_databases_sequences(encrypted_rds_instance)
            self.check_data_consistency(encrypted_rds_instance)
        else:
            self.logger.warning("One or more tasks finished with errors.")



import time
from datetime import UTC, datetime, timedelta
from typing import Optional

import boto3

from rds_encryptor.utils import get_logger


class ReplicationInstance:
    aws_client = boto3.client("dms")
    logger = get_logger("ReplicationInstance")

    def __init__(self, arn: str):
        self.arn = arn

    def get_status(self) -> str:
        response = self.aws_client.describe_replication_instances(
            Filters=[{"Name": "replication-instance-arn", "Values": [self.arn]}]
        )["ReplicationInstances"]
        if len(response) == 0:
            raise ValueError(f"Replication instance not found: {self.arn}")
        if len(response) > 1:
            raise ValueError(f"Multiple replication instances found: {self.arn}")

        return response[0]["ReplicationInstanceStatus"]

    def wait_until_active(self, timeout: int = 60 * 60, pooling_frequency: int = 60) -> "ReplicationInstance":
        timeout_dt = datetime.now(tz=UTC) + timedelta(seconds=timeout)
        self.logger.info('Waiting for replication instance "%s" to become active ...', self.arn)

        while datetime.now(tz=UTC) < timeout_dt:
            status = self.get_status()
            if status == "available":
                self.logger.info('Replication instance "%s" is active', self.arn)
                return self
            self.logger.debug("Replication instance %s is in status %s, waiting...", self.arn, status)
            time.sleep(pooling_frequency)

        raise TimeoutError(f"Replication instance {self.arn} creation timeout")

    @classmethod
    def from_arn(cls, arn: str) -> Optional["ReplicationInstance"]:
        assert arn, "Replication instance ARN is required"

        response = cls.aws_client.describe_replication_instances(
            Filters=[{"Name": "replication-instance-arn", "Values": [arn]}]
        )["ReplicationInstances"]
        if len(response) == 0:
            raise ValueError(f"Replication instance not found: {arn}")
        if len(response) > 1:
            raise ValueError(f"Multiple replication instances found: {arn}")

        return cls(arn=arn)



from threading import Thread

from rds_encryptor.dms.migration_task import MigrationTask, TaskFailedException
from rds_encryptor.utils import get_logger


class MigrationTaskManager:
    logger = get_logger("MigrationTaskManager")

    def __init__(self):
        self.tasks: list["MigrationTask"] = []
        self.errors = []

    def add_task(self, task: "MigrationTask"):
        self.tasks.append(task)

    def run_task(self, task: "MigrationTask"):
        try:
            self.logger.info('Starting database migration task "%s" ...', task.task_id)
            task.wait_until_ready().run_task().wait_until_finished()
            self.logger.info('Database migration task "%s" finished successfully', task.task_id)
        except TaskFailedException as e:
            self.errors.append(e)
            self.logger.error(
                'Database migration task "%s" with status %s because "%s" with last failure message "%s"',
                task.task_id,
                e.status,
                e.stop_reason,
                e.last_failure_message,
            )
        except TimeoutError as e:
            self.errors.append(e)
            self.logger.error(
                "[Task %s] Timeout error: %s. "
                "Task might be still running, if so, please increase timeout and try again.",
                task.task_id,
                e,
            )

    def run_all(self) -> bool:
        self.errors = []
        threads = []
        for task in self.tasks:
            thread = Thread(target=self.run_task, args=(task,))
            thread.start()
            threads.append(thread)
        for thread in threads:
            thread.join()

        return not self.errors
